import json
from typing import Union
from urllib.parse import urlencode

from discogs_client import models
from discogs_client.exceptions import ConfigurationError, HTTPError, AuthorizationError
from discogs_client.utils import update_qs
from discogs_client.fetchers import RequestsFetcher, OAuth2Fetcher, UserTokenRequestsFetcher


class Client:
    _base_url = 'https://api.discogs.com'
    _request_token_url = 'https://api.discogs.com/oauth/request_token'
    _authorize_url = 'https://www.discogs.com/oauth/authorize'
    _access_token_url = 'https://api.discogs.com/oauth/access_token'

    def __init__(self, user_agent, consumer_key=None, consumer_secret=None, token=None, secret=None, user_token=None):
        """An interface to the Discogs API."""
        self.user_agent = user_agent
        self.verbose = False
        self._fetcher = RequestsFetcher()

        if consumer_key and consumer_secret:
            self.set_consumer_key(consumer_key, consumer_secret)
            if token and secret:
                self.set_token(token, secret)
        elif user_token is not None:
            self._fetcher = UserTokenRequestsFetcher(user_token)

    def set_consumer_key(self, consumer_key, consumer_secret):
        self._fetcher = OAuth2Fetcher(consumer_key, consumer_secret)

    def set_token(self, token, secret):
        try:
            self._fetcher.store_token(token, secret)
        except AttributeError:
            raise ConfigurationError('You must first set the consumer key and secret.')

    def get_authorize_url(self, callback_url=None):
        """
        Returns a tuple of (<access_token>, <access_secret>, <authorize_url>).
        Send a Discogs user to the authorize URL to get the verifier for the access token.
        """
        # Forget existing tokens
        self._fetcher.forget_token()

        params = {}
        params['User-Agent'] = self.user_agent
        params['Content-Type'] = 'application/x-www-form-urlencoded'
        if callback_url:
            params['oauth_callback'] = callback_url
        postdata = urlencode(params)

        content, status_code = self._fetcher.fetch(self, 'POST', self._request_token_url, 
                                                data=postdata, headers=params, json_format=False)
        if status_code != 200:
            raise AuthorizationError('Could not get request token.', status_code, content)

        token, secret = self._fetcher.store_token_from_qs(content)

        params = {'oauth_token': token}
        query_string = urlencode(params)

        return (token, secret, '?'.join((self._authorize_url, query_string)))

    def get_access_token(self, verifier):
        """
        Uses the verifier to exchange a request token for an access token.
        """
        self._fetcher.set_verifier(verifier)

        params = {}
        params['User-Agent'] = self.user_agent

        content, status_code = self._fetcher.fetch(self, 'POST', self._access_token_url, headers=params)
        if status_code != 200:
            raise HTTPError('Invalid response from access token URL.', status_code)

        token, secret = self._fetcher.store_token_from_qs(content)

        return token, secret

    def _check_user_agent(self):
        if not self.user_agent:
            raise ConfigurationError('Invalid or no User-Agent set.')

    def _request(self, method, url, data=None):
        if self.verbose:
            print(' '.join((method, url)))

        self._check_user_agent()

        headers = {
            'Accept-Encoding': 'gzip',
            'User-Agent': self.user_agent,
        }

        if data:
            headers['Content-Type'] = 'application/json'

        content, status_code = self._fetcher.fetch(self, method, url, data=data, headers=headers)

        if status_code == 204:
            return None

        body = json.loads(content)

        if 200 <= status_code < 300:
            return body
        else:
            raise HTTPError(body['message'], status_code)

    def _get(self, url):
        return self._request('GET', url)

    def _delete(self, url):
        return self._request('DELETE', url)

    def _post(self, url, data):
        return self._request('POST', url, data)

    def _patch(self, url, data):
        return self._request('PATCH', url, data)

    def _put(self, url, data):
        return self._request('PUT', url, data)

    def search(self, *query, **fields):
        """
        Search the Discogs database. Returns a paginated list of objects
        (Artists, Releases, Masters, and Labels). The keyword arguments to this
        function are serialized into the request's query string.
        """
        if query:
            items = [
                item.decode() if type(item) == bytes else item for item in query
            ]
            fields['q'] = ' '.join(items)

        return models.MixedPaginatedList(
            self,
            update_qs(self._base_url + '/database/search', fields),
            'results'
        )

    def artist(self, id):
        """Fetch an Artist by ID."""
        return models.Artist(self, {'id': id})

    def release(self, id):
        """Fetch a Release by ID."""
        return models.Release(self, {'id': id})

    def master(self, id):
        """Fetch a Master by ID."""
        return models.Master(self, {'id': id})

    def label(self, id):
        """Fetch a Label by ID."""
        return models.Label(self, {'id': id})

    def list(self, id):
        """Fetch a List by ID."""
        return models.List(self, {'id': id})

    def user(self, username):
        """Fetch a User by username."""
        return models.User(self, {'username': username})

    def listing(self, id):
        """Fetch a Marketplace Listing by ID."""
        return models.Listing(self, {'id': id})

    def order(self, id):
        """Fetch an Order by ID."""
        return models.Order(self, {'id': id})

    def fee_for(self, price, currency='USD'):
        """Calculate the fee for selling an item on the Marketplace."""
        resp = self._get('{0}/marketplace/fee/{1:.4f}/{2}'.format(self._base_url, price, currency))
        return models.Price(self, {'value': resp['value'], 'currency': resp['currency']})

    def identity(self):
        """Return a User object representing the OAuth-authorized user."""
        resp = self._get(self._base_url + '/oauth/identity')
        return models.User(self, resp)

    @property
    def backoff_enabled(self) -> bool:
        return self._fetcher.backoff_enabled

    @backoff_enabled.setter
    def backoff_enabled(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError("Backoff enabled toggle should be of type bool")
        self._fetcher.backoff_enabled = value

    @property
    def connection_timeout(self):
        """Return current client connection timeout"""
        return self._fetcher.connect_timeout

    @property
    def read_timeout(self):
        """Return current client read timeout"""
        return self._fetcher.read_timeout

    def set_timeout(self,
                    connect: Union[int, float] = 5,
                    read: Union[int, float] = 10) -> None:
        """Set request timeout parameters

        Parameters
        ----------
            connect : (Union[int, float], optional)
                Time in seconds after which connect will timeout. Defaults to 5.
            read : (Union[int, float], optional)
                Time in seconds after which request will timeout. Defaults to 10.
        """
        self._fetcher.connect_timeout = connect
        self._fetcher.read_timeout = read



from requests.api import request
from oauthlib import oauth1
import json
import os
import re
from discogs_client.utils import backoff
from urllib.parse import parse_qsl
from typing import Union


class Fetcher:
    """
    Base class for Fetchers, which wrap and normalize the APIs of various HTTP
    libraries.

    (It's a slightly leaky abstraction designed to make testing easier.)
    """
    backoff_enabled = True
    connect_timeout: Union[float, int, None] = None
    read_timeout: Union[float, int, None] = None

    def fetch(self, client, method, url, data=None, headers=None, json=True):
        """Fetch the given request

        Parameters
        ----------
        client : object
            Instantiated discogs_client.client.Client object.
        method : str
            HTTP method.
        url : str
            API endpoint URL.
        data : dict, optional
            data to be sent in the request's body, by default None.
        headers : dict, optional
            HTTP headers, by default None.
        json_format : bool, optional
            If True, an object passed with the "data" arg will be converted into
            a JSON string, by default True.

        Returns
        -------
        content : bytes
            as returned by Python "Requests"
        status_code : int
            as returned by Python "Requests"

        Raises
        ------
        NotImplementedError
            Is raised if a child class doesn't implement a fetch method.
        """
        raise NotImplementedError()

    @backoff
    def request(self, method, url, data, headers, params=None):
        return request(
            method=method, url=url, data=data,
            headers=headers, params=params,
            timeout=(self.connect_timeout, self.read_timeout)
        )


class LoggingDelegator:
    """Wraps a fetcher and logs all requests."""
    def __init__(self, fetcher):
        self.fetcher = fetcher
        self.requests = []

    @property
    def last_request(self):
        return self.requests[-1] if self.requests else None

    def fetch(self, client, method, url, data=None, headers=None, json=True):
        """Appends passed "fetcher" to a requests list and returns result of
        fetcher.fetch method"""
        self.requests.append((method, url, data, headers))
        return self.fetcher.fetch(client, method, url, data, headers, json)


class RequestsFetcher(Fetcher):
    """Fetches via HTTP from the Discogs API (unauthenticated)"""
    def fetch(self, client, method, url, data=None, headers=None, json=True):
        """
        Parameters
        ----------
        client : object
            Unused in this subclass.
        method : str
            HTTP method.
        url : str
            API endpoint URL.
        data : dict, optional
            data to be sent in the request's body, by default None.
        headers : dict, optional
            HTTP headers, by default None.
        json_format : bool, optional
            Unused in this subclass, by default True.

        Returns
        -------
        content : bytes
            as returned by Python "Requests"
        status_code : int
            as returned by Python "Requests"
        """
        resp = self.request(method, url, data=data, headers=headers)
        self.rate_limit = resp.headers.get(
                'X-Discogs-Ratelimit')
        self.rate_limit_used = resp.headers.get(
                'X-Discogs-Ratelimit-Used')
        self.rate_limit_remaining = resp.headers.get(
                'X-Discogs-Ratelimit-Remaining')
        return resp.content, resp.status_code


class UserTokenRequestsFetcher(Fetcher):
    """Fetches via HTTP from the Discogs API using User-token authentication"""
    def __init__(self, user_token):
        self.user_token = user_token

    def fetch(self, client, method, url, data=None, headers=None, json_format=True):
        """Fetch the given request on the user's behalf

        Parameters
        ----------
        client : object
            Unused in this subclass.
        method : str
            HTTP method.
        url : str
            API endpoint URL.
        data : dict, optional
            data to be sent in the request's body, by default None.
        headers : dict, optional
            HTTP headers, by default None.
        json_format : bool, optional
            If True, an object passed with the "data" arg will be converted into
            a JSON string, by default True.

        Returns
        -------
        content : bytes
            as returned by Python "Requests"
        status_code : int
            as returned by Python "Requests"
        """
        data = json.dumps(data) if json_format and data else data
        resp = self.request(
            method, url, data=data, headers=headers, params={'token':self.user_token}
        )
        self.rate_limit = resp.headers.get(
                'X-Discogs-Ratelimit')
        self.rate_limit_used = resp.headers.get(
                'X-Discogs-Ratelimit-Used')
        self.rate_limit_remaining = resp.headers.get(
                'X-Discogs-Ratelimit-Remaining')
        return resp.content, resp.status_code


class OAuth2Fetcher(Fetcher):
    """Fetches via HTTP + OAuth 1.0a from the Discogs API."""
    def __init__(self, consumer_key, consumer_secret, token=None, secret=None):
        self.client = oauth1.Client(consumer_key, client_secret=consumer_secret)
        self.store_token(token, secret)

    def store_token_from_qs(self, query_string):
        token_dict = dict(parse_qsl(query_string))
        token = token_dict[b'oauth_token'].decode('utf-8')
        secret = token_dict[b'oauth_token_secret'].decode('utf-8')
        self.store_token(token, secret)
        return token, secret

    def forget_token(self):
        self.store_token(None, None)

    def store_token(self, token, secret):
        self.client.resource_owner_key = token
        self.client.resource_owner_secret = secret

    def set_verifier(self, verifier):
        self.client.verifier = verifier

    def fetch(self, client, method, url, data=None, headers=None, json_format=True):
        """Fetch the given request on the user's behalf

        Parameters
        ----------
        client : object
            Unused in this subclass.
        method : str
            HTTP method.
        url : str
            API endpoint URL.
        data : dict, optional
            Data to be sent in the request's body, by default None.
        headers : dict, optional
            HTTP headers, by default None.
        json_format : bool, optional
            If True, an object passed with the "data" arg will be converted into
            a JSON string, by default True.

        Returns
        -------
        content : bytes
            as returned by Python "Requests"
        status_code : int
            as returned by Python "Requests"
        """
        body = json.dumps(data) if json_format and data else data
        uri, headers, body = self.client.sign(url, http_method=method,
                                              body=body, headers=headers)

        resp = self.request(method, url, data=body, headers=headers)
        self.rate_limit = resp.headers.get(
                'X-Discogs-Ratelimit')
        self.rate_limit_used = resp.headers.get(
                'X-Discogs-Ratelimit-Used')
        self.rate_limit_remaining = resp.headers.get(
                'X-Discogs-Ratelimit-Remaining')
        return resp.content, resp.status_code


class FilesystemFetcher(Fetcher):
    """Fetches from a directory of files."""
    default_response = json.dumps({'message': 'Resource not found.'}), 404
    path_with_params = re.compile('(?P<dir>(\w+/)+)(?P<query>\w+)\?(?P<params>.*)')

    def __init__(self, base_path):
        self.base_path = base_path

    def fetch(self, client, method, url, data=None, headers=None, json=True):
        """Fetch the given request

        Parameters
        ----------
        client : object
            Instantiated discogs_client.client.Client object.
        method : str
            Unused in this subclass (this fetcher supports GET only).
        url : str
            API endpoint URL.
        data : dict, optional
            Unused in this subclass (this fetcher supports GET only).
        headers : dict, optional
            Unused in this subclass.
        json_format : bool, optional

        Returns
        -------
        content : bytes
        status_code : int
        """
        url = url.replace(client._base_url, '')

        if json:
            base_name = ''.join((url[1:], '.json'))
        else:
            base_name = url[1:]

        path = os.path.join(self.base_path, base_name)

        # The exact path might not exist, but check for files with different
        # permutations of query parameters.
        if not os.path.exists(path):
            base_name = self.check_alternate_params(base_name, json)
            path = os.path.join(self.base_path, base_name)

        try:
            path = path.replace('?', '_')  # '?' is illegal in file names on Windows
            with open(path, 'r') as f:
                content = f.read().encode('utf8')  # return bytes not unicode
            return content, 200
        except:
            return self.default_response

    def check_alternate_params(self, base_name, json):
        """
        parse_qs() result is non-deterministic - a different file might be
        requested, making the tests fail randomly, depending on the order of parameters in the query.
        This fixes it by checking for matching file names with a different permutations of the parameters.
        """
        match = self.path_with_params.match(base_name)

        # No parameters in query - no match. Nothing to do.
        if not match:
            return base_name

        ext = '.json' if json else ''

        # The base name consists of one or more path elements (directories),
        # a query (discogs.com endpoint), query parameters, and possibly an extension like 'json'.
        # Extract these.
        base_dir = os.path.join(self.base_path, match.group('dir'))
        query = match.group('query')  # we'll need this to only check relevant filenames
        params_str = match.group('params')[:-len(ext)]  # strip extension if any
        params = set(params_str.split('&'))

        # List files that match the same query, possibly with different parameters
        filenames = [f for f in os.listdir(base_dir) if f.startswith(query)]
        for f in filenames:
            # Strip the query, the '?' sign (or its replacement) and the extension, if any
            params2_str = f[len(query) + 1:-len(ext)]
            params2 = set(params2_str.split('&'))
            if params == params2:
                return base_name.replace(params_str, params2_str)

        # No matching alternatives found - revert to original.
        return base_name


class MemoryFetcher(Fetcher):
    """Fetches from a dict of URL -> (content, status_code)."""
    default_response = json.dumps({'message': 'Resource not found.'}), 404

    def __init__(self, responses):
        self.responses = responses

    def fetch(self, client, method, url, data=None, headers=None, json=True):
        """Fetch the given request

        Parameters
        ----------
        client : object
            Unused in this subclass.
        method : str
            Unused in this subclass (this fetcher supports GET only).
        url : str
            API endpoint URL.
        data : dict, optional
            Unused in this subclass (this fetcher supports GET only).
        headers : dict, optional
            Unused in this subclass.
        json_format : bool, optional
            Unused in this subclass

        Returns
        -------
        content : bytes
        status_code : int
        """
        return self.responses.get(url, self.default_response)
